#!/usr/bin/python
# -*- coding: utf-8 -*-

import logging
import traceback

from octopus.core import op_codes

_KEY_NAME_SUFFIX = '_key'
_RECONCILIATION_KEY_LIST_DELIMITER = ','

# Config option names
_SOURCE_NAME_OPTION = 'source_name'
_KEY_NAME_OPTION = 'key_name'
_PRIMARY_KEY_OPTION = 'primary_key'
_RECONCILIATION_KEY_LIST_OPTION = 'reconciliation_keys'
_EXTENDED_DATA_FILE_OPTION = 'extended_data_file'
_EXTENDED_DATA_SET_NAME_OPTION = 'extended_data_set'


class RecoverableError(Exception):

    def __init__(self, message='Recoverable error', cause=None):
        self.message = message
        self.cause = cause


class InputError(Exception):

    def __init__(self, key_name, key_value, explanation='', cause=None):
        self.message = 'Bad input for record: {} = {}{}.'.format(key_name, key_value, explanation)
        self.key_name = key_name
        self.key_value = key_value
        self.cause = cause


class MissingRequiredFieldError(InputError):

    def __init__(self, key_name, key_value, missing_field_name):
        super().__init__(key_name, key_value, '-> Required field {} is missing'.format(missing_field_name))
        self.missing_field_name = missing_field_name


class ProcessingError(Exception):
    def __init__(self, key_name, key_value, explanation='', cause=None):
        self.message = 'Processing error for record: {} = {}. {}'.format(key_name, key_value, explanation)
        self.key_name = key_name
        self.key_value = key_value
        self.cause = cause


class DuplicateKeyError(ProcessingError):

    def __init__(self, key_name, key_value):
        super().__init__(key_name, key_value, '-> Duplicate key')


class DuplicateAltKeyError(ProcessingError):

    def __init__(self, key_name, key_value):
        super().__init__(key_name, key_value, '-> Duplicate alternative key')


class NonDeterministicCaseError(ProcessingError):
    """
        NonDeterministicCaseError is raised when the handler can't decide the behavior to execute. For instance,
        it happens if multiple persistent records match the lookup criteria when a PersistentHandler try to upsert
        a record.
    """

    def __init__(self, key_name, key_value, match_count: int, excluded_alt_key_name=None):
        super().__init__(key_name, key_value, '-> Non deterministic case: Found {} matches'.format(match_count))
        if excluded_alt_key_name is not None:
            self.message = self.message + ', with {} excluded'.format(excluded_alt_key_name)


class ExtendedData:
    # TODO: Use another type of source (e.g. CSV)

    def __init__(self, key_name, workbook_file=None, sheet_name=None):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.records = {}
        if workbook_file is not None and sheet_name is not None:
            from openpyxl import load_workbook
            wb = load_workbook(workbook_file)
            sheet = wb[sheet_name]
            row_count = sheet.max_row
            column_count = sheet.max_column
            column_heads = []
            for col_num in range(1, column_count + 1):
                column_heads.append(sheet.cell(1, col_num).value)
            for row_num in range(2, row_count + 1):
                record = {}
                for col_num in range(1, column_count + 1):
                    record[column_heads[col_num - 1]] = sheet.cell(row_num, col_num).value
                key_value: str = record.get(key_name, None)
                if key_value is not None:
                    if not isinstance(key_value, str):
                        self.records[record[key_name]] = record
                    elif not key_value.isspace():
                        self.records[record[key_name]] = record
                    else:
                        self.logger.debug('Ignored extended data entry: {}'.format(record))
                else:
                    self.logger.debug('Missing {} for data entry: {}'.format(key_name, record))

    def lookup_extended_data(self, key_value):
        return self.records.get(key_value, {})


class AbstractHandler:
    def __init__(self, config):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.source_name = config[_SOURCE_NAME_OPTION]
        self.key_name = config.get(_KEY_NAME_OPTION, self.source_name + _KEY_NAME_SUFFIX)
        extended_data_file = config.get(_EXTENDED_DATA_FILE_OPTION, None)
        extended_data_set_name = config.get(_EXTENDED_DATA_SET_NAME_OPTION, 'data')
        self.extended_data = ExtendedData(self.key_name, extended_data_file, extended_data_set_name)
        self.primary_key_name = config[_PRIMARY_KEY_OPTION]
        # TODO: Override op_logging_levels from config
        self.op_logging_levels = op_codes.OP_DEFAULT_LOGGING_LEVELS

    def handle(self, record, allow_retry=False):
        try:
            obj = self._build_extended_object(record)
            if obj is None:
                op_code = op_codes.IGNORE
            else:
                op_code = self._process_obj(obj)
        except RecoverableError as e:
            self.logger.error('{} - RECOVERABLE ERROR - {}'.format(self.source_name, e))
            if allow_retry:
                op_code = op_codes.RETRY
            else:
                op_code = op_codes.ERROR
        except InputError as e:
            self.logger.error('{} - INPUT ERROR - {}'.format(self.source_name, e.message))
            op_code = op_codes.ERROR
        except ProcessingError as e:
            self.logger.error('{} - PROCESSING ERROR - {}'.format(self.source_name, e.message))
            op_code = op_codes.ERROR
        except Exception as e:
            self.logger.error('{} - UNEXPECTED ERROR - {}'.format(self.source_name, e))
            op_code = op_codes.ERROR
            traceback.print_exc()
        self.logger.log(self.op_logging_levels.get(op_code, logging.ERROR),
                        '{} - {} - {}'.format(self.source_name, op_code, record))
        return op_code

    def _build_extended_object(self, record):
        raise NotImplemented()

    def _process_obj(self, obj):
        """
        Actual processing of a handled objet
        :param obj: The object to process
        :return: The executed operation code
        """
        raise NotImplemented()


class PersistenceService:

    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)

    def get_key_value(self, obj):
        raise NotImplemented()

    def lookup(self, obj):
        raise NotImplemented()

    def insert(self, obj):
        raise NotImplemented()

    def update(self, old_obj, new_obj):
        raise NotImplemented()

    def delete(self, obj):
        raise NotImplemented()


class PersistenceHandler(AbstractHandler):

    def __init__(self, config, service: PersistenceService):
        super().__init__(config)
        self.service = service
        reconciliation_key_list = config.get(_RECONCILIATION_KEY_LIST_OPTION, None)
        if reconciliation_key_list is None:
            self.reconciliation_keys = []
        else:
            self.reconciliation_keys = reconciliation_key_list.split(_RECONCILIATION_KEY_LIST_DELIMITER)

    def _process_obj(self, obj):
        """
        Actual processing consists on trying to performs an upsert (insert or update) of the handled object throught
        a persistence service.
        :param obj: The object to process
        :return: The executed operation code
        """
        matches = self.service.lookup(obj)
        match_count = len(matches)
        if match_count == 0:
            self.service.insert(obj)
            op_code = op_codes.INSERT
        elif match_count == 1:
            self.service.update(matches[0], obj)
            op_code = op_codes.UPDATE
        else:
            self.logger.debug('Found {} matches for object: {}'.format(match_count, obj))
            raise NonDeterministicCaseError(self.primary_key_name, self.service.get_key_value(obj), match_count)
        return op_code

    def _build_extended_object(self, record):
        raise NotImplemented()


class DummyDumpHandler(AbstractHandler):
    _OUTPUT_FILE = 'output_file'

    def __init__(self, config):
        super().__init__(config)
        output_file_path = config.get(DummyDumpHandler._OUTPUT_FILE, 'out.txt')
        self.output_file = open(output_file_path, 'w')

    def _build_extended_object(self, record):
        return record

    def _process_obj(self, obj):
        self.output_file.write(str(obj) + '\n')
        return op_codes.SUCCESS
